#ifndef LOGGER_V33_REF_H
#define LOGGER_V33_REF_H

#include <array>
#include <atomic>
#include <chrono>
#include <cstring>
#include <dlfcn.h>
#include <iostream>

namespace Details
{
template <std::size_t bufferSizePower2>
class CircularBuffer
{
public:
    static constexpr std::size_t bufferSize = 2 ^ bufferSizePower2;

    void clear()
    {
        m_nonModTail = 0u;
        m_bufferFull = false;
    }
    template <typename T>
    void write(const T t)
    {
        // NOTE: compiler will replace module power-of-2 by non-branching AND
        const std::size_t tail = m_nonModTail.fetch_add(sizeof(T)) % bufferSize;
        if (tail < bufferSize - sizeof(T))
        {
            // This hot path will typically not create t as such, but prefer a
            // direct copy from the input arguments
            std::memcpy(&m_buffer[tail], &t, sizeof(T));
        }
        else
        {
            m_bufferFull = true;
            const std::size_t firstPart = bufferSize - tail;
            const auto explicitT = t; // only here, the T will be constructed
            std::memcpy(&m_buffer[tail], &explicitT, firstPart);
            std::memcpy(&m_buffer[0], reinterpret_cast<const char *>(&explicitT) + firstPart, sizeof(T) - firstPart);
        }
    }

    void writeTo(std::ostream &s) const
    {
        const std::size_t tail = m_nonModTail % bufferSize;
        if (m_bufferFull)
        {
            s.write(&m_buffer[tail], static_cast<std::streamsize>(bufferSize - tail));
        }
        s.write(m_buffer.data(), static_cast<std::streamsize>(tail));
    }

private:
    std::array<char, bufferSize> m_buffer{};
    std::atomic<std::size_t> m_nonModTail{0u};
    bool m_bufferFull = false;
};

}; // namespace Details

template <std::size_t bufferSizePower2>
class Logger
{
    // Logging
public:
    template <typename... Ts>
    void __attribute__((always_inline)) trace(const Ts... args)
    {
        trace_inner(instructionpointer(), args...);
    }
    template <typename... Ts>
    inline void __attribute__((noinline, used)) trace_inner(const uintptr_t traceLocation, const Ts... args)
    {
        m_circularBuffer.write(Record<Ts...>{traceLocation, instructionpointer(), now(), args...});
    }

    static inline uintptr_t instructionpointer() __attribute__((always_inline))
    {
        uintptr_t traceLocation;
        asm volatile("lea (%%rip),%0" : "=r"(traceLocation));
        return traceLocation;
    }

    using Clock = std::chrono::high_resolution_clock;
    using TimeUnit = std::chrono::nanoseconds;
    using TimeRep = TimeUnit::rep;
    static inline TimeRep now()
    {
        return duration_cast<TimeUnit>(Clock::now().time_since_epoch()).count();
    }

private:
    struct __attribute__((packed)) RecordBase
    {
        const uintptr_t m_callerLocation;
        const uintptr_t m_traceLocation;
        const TimeRep m_time;
    };

    template <std::size_t I, typename T>
    struct __attribute__((packed)) RecordArg
    {
        T arg;
    };

    template <typename...>
    struct __attribute__((packed)) Record;
    template <std::size_t... Is, typename... Ts>
    struct __attribute__((packed)) Record<std::index_sequence<Is...>, Ts...> : RecordBase, RecordArg<Is, Ts>...
    {
        Record(uintptr_t callerLocation, uintptr_t traceLocation, const TimeRep time, const Ts... args)
            : RecordBase(callerLocation, traceLocation, time), RecordArg<Is, Ts>(args)...
        {
        }
    };
    template <typename... Ts>
    struct Record : Record<std::make_index_sequence<sizeof...(Ts)>, Ts...>
    {
        using Record<std::make_index_sequence<sizeof...(Ts)>, Ts...>::Record;
    };

    // Buffer
public:
    void writeTo(std::ostream &s) const
    {
        m_circularBuffer.writeTo(s);
    }

private:
    Details::CircularBuffer<bufferSizePower2> m_circularBuffer;
};

#endif // LOGGER_V33_REF_H
